package compiler;

import assem.Proc;

import error.ErrorMsg;
import parser.TigerParser;     // parser produced by CUP
import syntax.Exp;
import semantics.Semantics;

import java.io.*;
import java.util.*;

import translate.Fragment;
import translate.DataFrag;
import translate.ProcFrag;

import tree.StmList;

import frame.Frame;

import assem.Instr;

import regalloc.RegAlloc;

public class Main {

    private static final Frame frame = Compiler.frame;
    
    public static boolean verbose = System.getProperty ("verbose")!=null;

    public static void main (String args[]) throws IOException {
        if(args.length==0) {
            System.out.println("Usage: java -jar compiler.jar $INPUTFILE");
            return;
        }
        
        final String filename = args[0];
        final String flags = (args.length>1) ? args[1] : "";

        final int loc = filename.lastIndexOf('.');
        final String fileout = (loc!=-1) ?
            filename.substring(0, loc) + ".s" : filename + ".s";
        
        final Proc code = new Proc();

        System.out.println("Reading: " + filename);
        
        try {
            final ErrorMsg errorMsg = new ErrorMsg(filename);
            final InputStream is = new FileInputStream(filename);

            final TigerParser parser =  new TigerParser(is, errorMsg);
            final Exp e = (Exp) parser.parse().value;
            is.close();

            chkerr(errorMsg);
         
            final Semantics s = new Semantics(errorMsg);
            final LinkedList frags = s.transProg(e);

            chkerr(errorMsg);

            // complete code, in strings
            for(Iterator i=frags.iterator(); i.hasNext(); ) {
                final Fragment f = (Fragment)i.next();
                if(f instanceof DataFrag) {
                    code.append(f.toString());
                }
                else if(f instanceof ProcFrag)  {
                    ProcFrag frag = (ProcFrag)f;
                    Frame frame = frag.frame;

                    // add incoming/outgoing argument pseudo-code
                    frag.body = frame.procEntryExit1(frag.body);

                    // canonocalize
                    final StmList stms = canon.Main.transform(frag.body); 

                    RegAlloc ra = new RegAlloc(stms, frame);

                    // generate code
                    List instructions = frame.codegen(stms, ra);
                    
                    // add prologue/epilogue code
                    frame.procEntryExit3(instructions);
                    
                    for(Iterator ii=instructions.iterator(); ii.hasNext(); )
                        code.append((((Instr)ii.next()).format(frame)));
                }
                code.append(); // new line between fragments
            }
        }
        catch(FileNotFoundException e) {
            System.out.println("File not found: " + filename);
            System.exit(1);
        }
        catch(Exception e) {
            e.printStackTrace(System.err);
            System.exit(1);
        }

        frame.procEntryExit4(code);

        if(!flag(flags, 'n')) {
            System.out.println("Writing: " + fileout);
            PrintWriter fout =
                new PrintWriter(new FileOutputStream(fileout), true);
            fout.println(code);
        }

        if(flag(flags, 's')) System.out.println(code);
    }

    private static boolean flag(String flags, char flag) 
    { return flags.indexOf(flag)!=-1; }
    
    private static void chkerr(ErrorMsg em) 
    {
        if(em.getErrors()>0) {
            System.out.println(em.getErrors() + " errors.");
            System.exit(1);
        }
    }
}
