package semantics;

import error.ErrorMsg;
import compiler.Compiler;
import debug.Debug;
import symbol.*;
import syntax.*;
import translate.Fragment;
import translate.Translate;
import translate.Level;
import translate.Access;
import java.util.List;
import java.util.LinkedList;
import temp.Label;

public class Semantics {

    // used for return values that need both exps and types
    private class ExpTy {
        translate.Exp exp;
        Type type;
        ExpTy(translate.Exp e, Type t)
        { exp=e; type=t; }
    }

    // global error handler
    ErrorMsg errMsg;

    // used for checking break() legitimacy
    int forLoop = 0;
    int whileLoop = 0;
    
    // symbol tables
    SymbolTable types = new SymbolTable();
    SymbolTable vars  = new SymbolTable();
    SymbolTable funcs = new SymbolTable();
    
    // show all debugging information
    private static boolean verbose = System.getProperty("verbose")!=null;
    private int indent = 0;
  
    // all translation utilities
    Translate trans = new Translate();
    Level     level;
  

    // SAFE is for production because it prevents crashing on bad tiger code
    // UNSAFE is for testing because it helps find complier bugs
    final ExpTy ET_ERROR_SAFE = new ExpTy(trans.nop(), Type.ERROR);
    final ExpTy ET_ERROR_UNSAFE = new ExpTy(null, Type.ERROR);
    final ExpTy ET_ERROR = ET_ERROR_SAFE;

    // general constructor
    public Semantics(ErrorMsg err) {
        errMsg = err;
    }

    private LinkedList args() 
    { return new LinkedList(); }

    private LinkedList args(Object a)
    {
        LinkedList list = new LinkedList();
        list.add(a);
        return list;
    }
    
    private LinkedList args(Object a, Object b)
    {
        LinkedList list = new LinkedList();
        list.add(a);
        list.add(b);
        return list;
    }
    

    private static final String extfuncs[] =
    { "print", "print_int", "getch", "getstr", "ord", "chr", "flush" };

    private static boolean isExtFunc(String s) 
    {
        for(int i=0; i<extfuncs.length; i++)
            if(extfuncs[i].equals(s))
                return true;
        return false;
    }
    
    // general translater
    public Fragment transProg(Exp exp) {
        if(verbose) Debug.on(System.out);
        else Debug.off();

        beginScope("global");
        beginLevel(new Label(Compiler.ENTRYPOINT, true));

        // add predefined types to scope table
        defType(Type.STRING.toString(), Type.STRING);
        defType(Type.INT.toString(), Type.INT);
        defType(Type.NIL.toString(), Type.NIL);
        defType(Type.VOID.toString(), Type.VOID);

        final Type STRING = Type.STRING;
        final Type VOID   = Type.VOID;
        final Type INT    = Type.INT;
        
        defFunc("print",  args(STRING), VOID,   new Label("print", true));
        defFunc("print_int", args(INT), VOID,   new Label("print_int", true));
        defFunc("getch",  args(),       STRING, new Label("getch", true));
        defFunc("getstr", args(),       STRING, new Label("getstr", true));
        defFunc("ord",    args(STRING), INT,    new Label("ord", true));
        defFunc("chr",    args(INT),    STRING, new Label("chr", true));
        defFunc("flush",  args(),       VOID,   new Label("flush", true));

        trans.procEntryExit(level, transExp(exp).exp);

        endLevel();
        endScope();

        return trans.getFragments();
    }
  
    // very generic dispatch unit
    private ExpTy trans(Object o) {
        if(o instanceof Exp) return transExp((Exp)o);
        if(o instanceof Var) return transVar((Var)o);
        if(o instanceof Dec) return transDec((Dec)o);
        return ET_ERROR;
    }

    /*
      EXPRESSIONS
    */
    private ExpTy transExp(Exp exp) {
        if(exp instanceof AssignExp) return transExp((AssignExp)exp);
        if(exp instanceof BreakExp)  return transExp((BreakExp)exp);
        if(exp instanceof CallExp)   return transExp((CallExp)exp);
        if(exp instanceof WhileExp)  return transExp((WhileExp)exp);;
        if(exp instanceof ForExp)    return transExp((ForExp)exp);
        if(exp instanceof IfExp)     return transExp((IfExp)exp);
        if(exp instanceof IntExp)    return transExp((IntExp)exp);
        if(exp instanceof LetExp)    return transExp((LetExp)exp);
        if(exp instanceof NilExp)    return transExp((NilExp)exp);
        if(exp instanceof OpExp)     return transExp((OpExp)exp);
        if(exp instanceof ArrayExp)  return transExp((ArrayExp)exp);
        if(exp instanceof RecordExp) return transExp((RecordExp)exp);
        if(exp instanceof SeqExp)    return transExp((SeqExp)exp);
        if(exp instanceof StringExp) return transExp((StringExp)exp);;
        if(exp instanceof VarExp)    return transExp((VarExp)exp);
        e_UnknownSubclass(exp.pos, "transExp  - ", exp);
        return new ExpTy(null, Type.VOID);  
    }

    private ExpTy transExp(AssignExp exp) {
        ExpTy etlhs = transVar(exp.var);
        ExpTy etrhs = transExp(exp.exp);

        if(etlhs==ET_ERROR || etrhs==ET_ERROR)
            return ET_ERROR;
    
        Type tlhs = actual(etlhs.type);
        Type trhs = actual(etrhs.type);
        if(!trhs.coerceTo(tlhs) && !(tlhs==Type.ERROR))
            e_IncompatibleTypes(exp.pos, tlhs, trhs);
        return new ExpTy(assign(etlhs.exp, etrhs.exp), Type.VOID);
    }

    private ExpTy transExp(BreakExp exp) {
        if(forLoop==0 && whileLoop==0)
            error(exp.pos, "Break outside of a for- or while-loop.");
    
        return new ExpTy(null, Type.VOID);
    }

    private ExpTy transExp(CallExp exp) {
        if(!funcExists(exp.func))
            e_UndefinedFunction(exp.pos, exp.func);
        else {
            // function exists, check args
            FuncEntry fe = getFuncEntry(exp.func);
            LinkedList fargs = fe.args;  // actual function args <Type>
            LinkedList eargs = exp.args; // args in expression   <Exp>
            LinkedList targs = new LinkedList();  // trans.funCall()<trans.Exp>
            if(fargs.size()!=eargs.size())
                error(exp.pos, "Incorrect number of arguments.");
            else {
                boolean fErr = false;
        
                for(int i=0; i<eargs.size(); i++) {
                    Type fargtype = (Type)fargs.get(i); 
                    Exp  earg = (Exp)eargs.get(i);
                    ExpTy etearg = transExp(earg);
                    Type eargtype = etearg.type;
          
                    if(etearg==ET_ERROR)
                        fErr=true;            
            
                    targs.add(etearg.exp);
                    if(!eargtype.coerceTo(fargtype)) {
                        error(exp.pos,
                              "Function parameter #"+(i+1)+
                              " is incorrect type:\n" +
                              "  Function: " + exp.func + "\n" +
                              "  Found:    " + eargtype + "\n" +
                              "  Expected: " + fargtype);
                        fErr = true;
                    }
                }
                if(!fErr) {
                    boolean bStatic = ! isExtFunc(fe.label.toString());
                    return new ExpTy(trans.call(level, fe.level, fe.label,
                                                targs, bStatic),
                                     fe.result);
                }
                else
                    return ET_ERROR;
            }
        }
        return ET_ERROR;
    }
  
    private ExpTy transExp(ForExp exp) {
        beginScope("for");
        forLoop++;
    
        Type tvar  = actual(transDec(exp.var).type);
        Type tinit = actual(transExp(exp.var.init).type);
        Type thigh = actual(transExp(exp.hi).type);
        Type tbody = actual(transExp(exp.body).type);

        if(tinit!=Type.INT)
            error(exp.pos, "For loop: iterator must be an integer.");

        if(thigh!=Type.INT)
            error(exp.pos, "For loop: high value must be an integer.");
    
        /*
          turn for(...) into
          let
          var v:=init
          in
          while(v<hi)
          body
          v++
          end
        */

        int p = exp.pos;
        // iterator
        Var var = new SimpleVar(p, exp.var.name);

        // these need to be read in the order of execution, or probs will arise
        LinkedList decs  = new LinkedList(); // declarations for let
        decs.add(exp.var);
    
        // generate the while loop (execute body + increment counter)
        LinkedList wbody = new LinkedList();
        Exp winc = new AssignExp(p, var,
                                 new OpExp(p, new VarExp(p, var),
                                           OpExp.PLUS,
                                           new IntExp(p, 1)));
        wbody.add(exp.body);
        wbody.add(winc);

        WhileExp whileexp = new WhileExp(p, new OpExp(p, new VarExp(p, var),
                                                      OpExp.LE,
                                                      exp.hi),
                                         new SeqExp(p, wbody));

        LinkedList body  = new LinkedList();
        body.add(whileexp);
    
        ExpTy etlet = transExp(new LetExp(p, decs, body));

        forLoop--;
        endScope();
    
        return new ExpTy(etlet.exp, Type.VOID);
    }

    private ExpTy transExp(WhileExp exp) {
        whileLoop++;

        ExpTy ettest = transExp(exp.test);
        ExpTy etbody = transExp(exp.body);
        Type ttest = actual(ettest.type);

        if(ttest!=Type.INT)
            error(exp.pos, "While loop: test must return an integer.");

        whileLoop--;
        return new ExpTy(trans.whileExp(ettest.exp, etbody.exp), Type.VOID);
    }
  
    private ExpTy transExp(IfExp exp) {
        ExpTy ettest = transExp(exp.test);
        ExpTy etthen = transExp(exp.thenclause);

        ExpTy etelse = (exp.elseclause==null) ? null:transExp(exp.elseclause);
    
        Type ttest = actual(ettest.type);
        Type tthen = actual(etthen.type);
        Type telse = (etelse==null) ? null : actual(etelse.type);

        if(ttest != Type.INT)
            error(exp.pos, "If loop: test must return an integer");
        if(telse!=null)
            if(tthen.coerceTo(telse) ||
               ((tthen instanceof Record) && (telse==Type.NIL)) ||
               ((telse instanceof Record) && (tthen==Type.NIL))) {
                /* valid */
            }
            else
                error(exp.pos,
                      "THEN- and ELSE-clauses must return the same type.\n" +
                      "  <THEN>: " + tthen + "\n" +
                      "  <ELSE>: " + telse);
        translate.Exp etest = ettest.exp;
        translate.Exp ethen = etthen.exp;
        translate.Exp eelse = (etelse==null) ? null : etelse.exp;    
        return new ExpTy(trans.ifExp(etest, ethen, eelse), tthen);
    }
  
    private ExpTy transExp(IntExp exp)
    { return new ExpTy(trans.constInt(exp.value), Type.INT); }
  
    private ExpTy transExp(NilExp exp)
    { return new ExpTy(trans.nil(), Type.NIL); }
  
    private ExpTy transExp(StringExp exp)
    { return new ExpTy(trans.string(level, exp.value), Type.STRING); }

    private ExpTy transExp(OpExp exp) {
        ExpTy lhs  = transExp(exp.left);
        ExpTy rhs  = transExp(exp.right);
        Type  tlhs = actual(lhs.type);
        Type  trhs = actual(rhs.type);

        if(tlhs.coerceTo(trhs) ||
           ((tlhs instanceof Record) && (trhs==Type.NIL)) ||
           ((trhs instanceof Record) && (tlhs==Type.NIL))) {
            /* valid */
        }
        else {
            e_IncompatibleTypes(exp.pos, tlhs, trhs);
            return ET_ERROR;
        }

        // most common types, so set the flags now  
        boolean fInt = (tlhs == Type.INT);
        boolean fStr = (tlhs == Type.STRING);

        Type ret = null;
    
        switch(exp.oper) {
        case OpExp.PLUS: // strings will concatenate
        case OpExp.LT:   // strings will be compared lexicographically
        case OpExp.LE:
        case OpExp.GT:
        case OpExp.GE:
            if(!fInt && !fStr)
                error(exp.pos,
                      "Expressions must be either integers or strings.");
            else
                ret = (exp.oper==OpExp.PLUS) ? tlhs : Type.INT;
            break;
      
        case OpExp.MINUS:
        case OpExp.TIMES:
        case OpExp.DIV:
            if(!fInt)
                error(exp.pos, "Expressions must be integers.");
            else
                ret = tlhs;
            break;

            // all types are valid for eq/neq
        case OpExp.EQ:
        case OpExp.NE:
            ret = Type.INT;
        }

        // checks passed, now create code
        if(ret==null)
            return ET_ERROR;
        else {
            translate.Exp elhs = lhs.exp;
            translate.Exp erhs = rhs.exp;
            int op = -1;
            boolean f = false;

            switch(exp.oper) {
            case OpExp.PLUS:  op = Translate.PLUS;  f = true; break;
            case OpExp.MINUS: op = Translate.MINUS; f = true; break;
            case OpExp.TIMES: op = Translate.TIMES; f = true; break;
            case OpExp.DIV:   op = Translate.DIV;   f = true; break;
            }
            if(f)
                if(exp.oper==OpExp.PLUS && fStr)
                    return new ExpTy(trans.strcat(elhs, erhs), ret);
                else
                    return new ExpTy(trans.binop(op, elhs, erhs), ret);

            switch(exp.oper) {  
            case OpExp.LT: op = Translate.LT; f = true; break;
            case OpExp.LE: op = Translate.LE; f = true; break;
            case OpExp.GT: op = Translate.GT; f = true; break;
            case OpExp.GE: op = Translate.GE; f = true; break;
            case OpExp.EQ: op = Translate.EQ; f = true; break;
            case OpExp.NE: op = Translate.NE; f = true; break;
            }
            if(f)
                if(fStr)
                    return new ExpTy(trans.strcmp(op, elhs, erhs), ret);
                else // must be ints or EQ/NEQ
                    return new ExpTy(trans.cmpop(op, elhs, erhs), ret);

            return new ExpTy(null, ret);
        }
    }

    // array[count] of init
    private ExpTy transExp(ArrayExp exp) {
        // check if type exists
        Type  type   = actual(getType(exp.type));
        ExpTy etsize = transExp(exp.size);
        ExpTy etinit = transExp(exp.init);
        Type  tsize  = actual(etsize.type);
        Type  tinit  = actual(etinit.type);

        if(etsize==ET_ERROR || etinit==ET_ERROR)
            return ET_ERROR;
    
        if(!tsize.coerceTo(Type.INT))
            error(exp.pos, "Array size must be an integer");      
    
        if(type==null)
            e_UndefinedType(exp.pos, exp.type);
        else
            if(!(type instanceof Array)) // check if array
                error(exp.pos,
                      "Invalid type. " + exp.type + " is not an Array.");
            else // check if init type matches array type
                if(!tinit.coerceTo(actual(((Array)type).element)))
                    e_IncompatibleTypes(exp.pos,
                                        actual(((Array)type).element), tinit);
                else // success
                    return new ExpTy(trans.array(etsize.exp, etinit.exp),
                                     new Array(tinit));
        return ET_ERROR;
    }

    private ExpTy transExp(RecordExp exp) {
        Type type = actual(getType(exp.type));
        if(type==null)
            e_UndefinedType(exp.pos, exp.type);
        else 
            if(!(type instanceof Record)) // check if record
                error(exp.pos,
                      "Invalid type. " + exp.type + " is not a Record.");
            else {
                // check if all fields match correct types

                Record record = (Record)type;

                List recfields = record.fields;
                List expfields = exp.fields;
                // initial exps to send to trans
                LinkedList trinits = new LinkedList(); 
                // iterate through each field in expression
                // and make sure equivalent
                if(recfields.size()!=expfields.size())
                    error(exp.pos, "Invalid record length.");
                else {
                    boolean ferr = false;
                    for(int i=0; i<expfields.size(); i++) {
                        Record.Field recfield = (Record.Field)recfields.get(i);
                        VarDec       expfield = (VarDec)expfields.get(i);
                        Type rtype = recfield.fieldType;
                        if(!expfield.name.equals(recfield.fieldName)) {
                            error(exp.pos, "Field #" + (i+1) +
                                  " does not match with record.");
                            ferr = true;
                        }
                        else {
                            ExpTy etinit = transExp(expfield.init);
                            Type itype = actual(etinit.type);
                            if(!itype.coerceTo(rtype)) {
                                e_IncompatibleTypes(exp.pos, rtype, itype);
                                ferr=true;
                            }
                            trinits.add(etinit.exp);
                        }
                    }
                    // done with checking, if no errors, return type of Record
                    if(ferr)
                        return ET_ERROR;
                    else
                        return new ExpTy(trans.record(trinits), record);
                }
            }
        return ET_ERROR;
    }
  
    private ExpTy transExp(LetExp exp) {
        beginScope("let");
        ExpTy decs = transList(exp.decs);
        ExpTy body = transList(exp.body);
        endScope();
        // last type will be used
        return new ExpTy(trans.let(decs.exp, body.exp), body.type);
    }
  
    private ExpTy transExp(VarExp exp) {
        return transVar(exp.var);
    }

    private ExpTy transExp(SeqExp exp) {
        return transList(exp.list);
    }
  
    /*
      VARIABLES
    */
    private ExpTy transVar(Var var) {
        if(var instanceof SimpleVar)    return transVar((SimpleVar)var);
        if(var instanceof FieldVar)     return transVar((FieldVar)var);
        if(var instanceof SubscriptVar) return transVar((SubscriptVar)var);
        e_UnknownSubclass(var.pos, "transVar  - ", var);
        return ET_ERROR;
    }

    private ExpTy transVar(SimpleVar var) {
        if(varExists(var.name)) {
            VarEntry ve = getVarEntry(var.name);
            return new ExpTy(trans.simpleVar(level, ve.access),
                             actual(ve.type));
        }
        else{
            e_UndefinedVariable(var.pos, var.name);
            return ET_ERROR;
        }
    }

    private ExpTy transVar(SubscriptVar var) {
        ExpTy etvar = transVar(var.var);
        ExpTy etidx = transExp(var.index);
        if(!(etidx.type.coerceTo(Type.INT)))
            error(var.pos, "Index must be an integer.");
        if(!(actual(etvar.type) instanceof Array))
            error(var.pos, "Base variable is not an Array: " + name(var.var));
        else
            return new ExpTy(trans.subscriptVar(etvar.exp, etidx.exp),
                             ((Array)(etvar.type)).element );
        return ET_ERROR;
    }

    private ExpTy transVar(FieldVar var) {
        ExpTy etvar = transVar(var.var);
        if(!(actual(etvar.type) instanceof Record))
            error(var.pos, "Base variable is not a Record: " + name(var.var));
        else {
            List fields = ((Record)etvar.type).fields;
            for(int i=0; i<fields.size(); i++) {
                Record.Field field = (Record.Field)fields.get(i);
                if(field.fieldName.equals(var.field))
                    return new ExpTy(trans.fieldVar(etvar.exp, i),
                                     field.fieldType);
            }
            error(var.pos, "Field not found: " + var.field);
        }
        return ET_ERROR;
    }
  
    /*
      DECLARATIONS
    */
    private ExpTy transDec(Dec dec) {
        if(dec instanceof VarDec)      return transDec((VarDec)dec);
        if(dec instanceof TypeDec)     return transDec((TypeDec)dec);
        if(dec instanceof FunctionDec) return transDec((FunctionDec)dec);
        e_UnknownSubclass(dec.pos, "transDec  - ", dec);
        return ET_ERROR;
    }

    private ExpTy transDec(VarDec dec) {
        // make sure stated type and type of initial data are the same
        ExpTy etinit = transExp(dec.init);
        Type type    = actual(getType(dec.type));
        Type itype   = actual(etinit.type);
        boolean ferr = false;

        if(etinit == ET_ERROR)
            return ET_ERROR;
    
        if(dec.type!=null)
            if(type==null) {
                e_UndefinedType(dec.pos, dec.type);
                ferr = true;
            }
            else if(!itype.coerceTo(type)) {
                e_IncompatibleTypes(dec.pos, type, itype);
                ferr = true;
            }
            else
                defVar(dec.name, type);
        else
            if(itype==Type.NIL) {
                error(dec.pos,"Type must be specified for null initial value");
                ferr = true;
            }
            else
                defVar(dec.name, itype);

        if(ferr)
            return ET_ERROR;
        translate.Exp var = transVar(new SimpleVar(-1, dec.name)).exp;
        return new ExpTy(assign(var, etinit.exp), type.VOID);
    }

    private ExpTy transDec(TypeDec dec) {
        // first define all type names, then go through and define their types
        // assign to themselves to be able to check if its an error
        for(TypeDec cdec=dec; cdec!=null; cdec=cdec.next)
            defType(cdec.name, new Name(cdec.name));

        // go through again, and assign all type declarations
        for(TypeDec cdec=dec; cdec!=null; cdec=cdec.next)
            ((Name)getType(cdec.name)).bind(transType(cdec.type));

        // now check for loops
        for(TypeDec cdec=dec; cdec!=null; cdec=cdec.next) {
            Type type = getType(cdec.name);
            if(type==null)
                fatal_error("Expected type not found: " + cdec.name);
            else
                if((type instanceof Name) && (((Name)type).isLoop()))
                    error(cdec.pos, "Type " + cdec.name + " defines itself.");
        }        
        return new ExpTy(trans.nil(), Type.VOID);
    }

    private ExpTy transDec(FunctionDec dec) {

        // to allow recursion, first define all functions with their arguments
        // then go through and check all function bodies
      
        for(FunctionDec cdec=dec; cdec!=null; cdec=cdec.next) {
            Label label = new Label("f_" + cdec.name);
            
            beginScope(label);

            // this level won't be used but needs to be declared, so variables
            // don't get allocated in the parent
            beginLevel(label);

            LinkedList args = new LinkedList();
            // check parameters
            for(int i=0; i<cdec.params.size(); i++) {
                VarDec arg = (VarDec)cdec.params.get(i);
                Type argtype = getType(arg.type);
                if(argtype==null) {
                    e_UndefinedType(cdec.pos, arg.type);
                    argtype = Type.VOID; // prevent cause an undef var error
                }
                args.add(argtype);
                defVar(arg.name, argtype, true);
            }
      
            // check result type
            // restype will only be null if cdec.result exists, but the
            //   corresponding type is not defined
            Type restype = (cdec.result==null)?Type.VOID:getType(cdec.result);
            if(restype==null)
                e_UndefinedType(cdec.pos, cdec.result);

            endLevel();
            endScope();
            defFunc(cdec.name, args, restype, label);
        }

        // all functions have been defined, so now go through and check expr
        // need to define all variables again, but don't display var errors
        for(FunctionDec cdec=dec; cdec!=null; cdec=cdec.next) {
            // use initially defined label; used in recursive function calls 
            Label label = getFuncEntry(cdec.name).label;

            beginScope(label);
            beginLevel(label);

            LinkedList args = new LinkedList();
            // check parameters
            for(int i=0; i<cdec.params.size(); i++) {
                VarDec arg = (VarDec)cdec.params.get(i);
                Type argtype = getType(arg.type);
                if(argtype==null)
                    argtype = Type.VOID; // prevent cause an undef var error
                args.add(argtype);
                defVar(arg.name, argtype, true);
            }
            // check result type
            Type restype = (cdec.result==null)?Type.VOID:getType(cdec.result);
            
            // check function result matches reult type, also type-check body
            ExpTy etbody = transExp(cdec.body);
            Type bodytype = actual(etbody.type);
            if(restype!=Type.VOID && !bodytype.coerceTo(restype))
                error(cdec.pos,
                      "Function type is not equal to actual result type.");

            trans.procEntryExit(level, etbody.exp);
      
            endLevel();
            endScope();
        }

        return new ExpTy(trans.nil(), Type.VOID);
    }
  
    /*
      TYPE-DEFINES
    */
    private Type transType(TypeX type) {
        if(type instanceof NameType)   return transType((NameType)type);
        if(type instanceof ArrayType)  return transType((ArrayType)type);
        if(type instanceof RecordType) return transType((RecordType)type);
        return Type.ERROR;
    }

    private Type transType(NameType type) {
        // type <type1> = <type2>
        if(typeExists(type.name)) {
            Name name = new Name(type.name);
            name.bind(getType(type.name));
            return name;
        }
        else {
            e_UndefinedType(type.pos, type.name);
            return Type.ERROR;
        }
    }

    private Type transType(ArrayType arraytype) {
        // type <type> = array of <element>
        if(typeExists(arraytype.type)) { 
            Array array = new Array(getType(arraytype.type));
            return array;
        }
        else {
            e_UndefinedType(arraytype.pos, arraytype.type);
            return Type.ERROR;
        }
    }

    private Type transType(RecordType recordtype) {
        List fields = new LinkedList();
        for(int i=0; i<recordtype.fields.size(); i++) {
            VarDec dec = (VarDec)recordtype.fields.get(i);

            Type type = getType(dec.type);
            if(type==null)
                e_UndefinedType(recordtype.pos, dec.type);
            else
                if(type==Type.VOID)
                    error(recordtype.pos, "Field type may not be VOID.");
                else
                    fields.add(new Record.Field(dec.name, getType(dec.type)));
        }
        return new Record(fields);
    }
  
    /*
      LINKED LIST (for parameters, body, etc)
    */
    private ExpTy transList(LinkedList list) {
        ExpTy et = null;
        for(int i=0; i<list.size(); i++) {
            ExpTy cex = trans(list.get(i));
            // skip errors
            if(cex==ET_ERROR)
                continue;

            if(et==null)
                et = cex;
            else {
                et.exp = trans.eseq(et.exp, cex.exp);
                et.type = cex.type;
            }
        }
        if(et==null)
            return new ExpTy(trans.nop(), Type.VOID);
        else
            return et;
    }
  
    // general private routines
    private boolean varExists(String var)   { return getVarEntry(var)!=null;  }
    private boolean typeExists(String type) { return getTypeEntry(type)!=null;}
    private boolean funcExists(String func) { return getFuncEntry(func)!=null;}

    // just a wrapper for Type.actual()
    private Type actual(Type t) {
        return (t!=null) ? t.actual() : null;
    }

    private void defVar(String var, Type type)
    { defVar(var, type, false); }
  
    private void defVar(String var, Type type, boolean formal) 
    {
        Access access =  formal ? level.allocFormal() : level.allocLocal();
        vars.put(var, new VarEntry(type, access));
        debug("define var " + var);
        debug_level();
        debug_loc(access);
    }

    private void defType(String typename, Type type) 
    {
        debug("define type " + typename);
        debug_level();
        debug_loc();
        types.put(typename, new TypeEntry(type));
    }

    private void defFunc(String function, LinkedList args, Type result,
                         Label label) 
    {
        debug("define function " + function + ":" + result);
        debug_level();
        debug_loc();
        funcs.put(function, new FuncEntry(args, result, level, label));
    }  

    private Type getType(String type) {
        TypeEntry te = getTypeEntry(type);
        if(te!=null)
            return te.type;
        else
            return null;
    }
                          
    private VarEntry getVarEntry(String var) {
        if(var==null) return null;
        Entry e = (Entry)vars.get(var);
        if(e instanceof VarEntry) {
            debug("use var " + var);
            debug_level();
            debug_loc(((VarEntry)e).access);
            return (VarEntry)e;
        }
        else
            return null;
    }

    private TypeEntry getTypeEntry(String type) {
        if(type==null) return null;
        Entry e = (Entry)types.get(type);
        if(e instanceof TypeEntry) {
            /*
              debug("use type " + type);
              debug_level();
              debug_level(e.level);
              debug_loc(e.level);
            */
            return (TypeEntry)e;
        }
        else
            return null;
    }
  
    private FuncEntry getFuncEntry(String func) {
        if(func==null) return null;
        Entry e = (Entry)funcs.get(func);
        if(e instanceof FuncEntry) {
            debug("call function " + func);
            /*
              debug_level();
              debug_level(e.level);
              debug_loc(e.level);
            */
            return (FuncEntry)e;
        }
        else
            return null;
    }  

    // general scope functions

    private void beginScope(Label label) 
    { beginScope(label.label); }
    
    private void beginScope(String s) {
        debug("begin scope: " + s);
        indent++;
        types.beginScope();
        vars.beginScope();
        funcs.beginScope();
    }

    private void endScope() {
        indent--;
        debug("end scope");
        types.beginScope();
        vars.endScope();
        funcs.endScope();
    }

    public void beginLevel()         { level = new Level(level);    }
    public void beginLevel(String s) { level = new Level(level, s); }
    public void beginLevel(Label l)  { level = new Level(level, l); }
    public void endLevel()           { level = level.parent;        }

    // var should be returned from t.simpleVar(), t.subscriptVar(), etc
    public translate.Exp assign(translate.Exp var, translate.Exp exp)
    { return trans.assign(var, exp); }
  
    /*
      private void beginLevel(String s) { level = trans.beginLevel(s); }
      private void beginLevel() { level = trans.beginLevel(); }
      private void endLevel()   { level = trans.endLevel(); }
    */

    private void debug(String s) {
        for(int i=0; i<indent; i++)
            Debug.print("  ");
        Debug.println(s);
    }

    // random debugging functions
    private void debug_level()
    { /* debug("  current level: " + level+ " fp: " + level.fp()); */ }

    private void debug_loc()
    { debug_loc(level); }

    private void debug_level(Level level)
    { /* debug("  level of var:  " + level + " fp: " + level.fp()); */ }
  
    private void debug_loc(Level level)
    { /* debug("  loc of obj:    fp: " + level.fp() ); */ }

    private void debug_loc(Access access)
    {
        /*
        int cfp = level.fp();       // current frame pointer
        int vfp = access.home.fp(); // variable home frame pointer
        int vof = access.offset;    // variable offset in home frame
        int vex = vfp + vof;        // exact address of variable
        int cof = vof + vfp - cfp;  // variable offset from current frame ptr
    
        debug("  loc of var:    " + access.home + " fp: " + vfp +
              " offset: " + vof + " exact: " + vex + " off/cur: " + cof);
        */
    }
  
    // COMMON ERROR-REPORTERS
    private void e_UndefinedVariable(int pos, String var)
    { error(pos, "Undefined variable: " + var); }

    private void e_UndefinedType(int pos, String type)
    { error(pos, "Undefined type: " + type); }

    private void e_UndefinedFunction(int pos, String func)
    { error(pos, "Undefined function: " + func); }
  
    private void e_IncompatibleTypes(int pos, Type lhs, Type rhs)
    { error(pos, "Incompatible types:\n" +
            "  Found:    " + rhs + "\n" +
            "  Expected: " + lhs); }

    private void e_UnknownSubclass(int pos, String function, Object o)
    { error(pos, function + " - Unknown variable subclass:"
            + o.getClass().getName());  }
  
    private void error(int pos, String msg) 
    { errMsg.error(pos, msg); }

    // this is fatal because it should -never- be called.
    private void fatal_error(String msg) {
        System.out.println("FATAL ERROR: " + msg);
    }

    // return the complete name of a variable
    private String name(Var var) 
    {
        if(var instanceof SimpleVar)     return name((SimpleVar)var);
        if(var instanceof FieldVar)      return name((FieldVar)var);
        if(var instanceof SubscriptVar)  return name((SubscriptVar)var);
        return "";
    }

    private String name(SimpleVar var) 
    { return var.name;  }

    private String name(FieldVar var)
    { return name(var.var) + "." + var.field; }

    private String name(SubscriptVar var)
    { return name(var.var) + "[idx]"; }

    private void pause() 
    {
        try { System.in.read(); }
        catch(java.io.IOException e) {}
    }
}
